package grobs;

import java.awt.geom.Point2D ;
import java.util.* ;

import monitor.* ;

/**
 * <p>Title: Turtle World</p>
 * <p>Description: Turtle graphics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Theodore Norvell
 * @version 1.0
 */

public abstract class DynamicGrob extends Grob {

    protected final Monitor monitor = new Monitor() ;

    protected State state = new State() ;

    protected Thread executionThread ;

    protected List<Condition> conditionList = new ArrayList<Condition>() ;

    public DynamicGrob() {
        reset() ;
        executionThread = new UpdateThread() ;
        executionThread.start(); }

    protected void addCondition( Condition condition ) {
        conditionList.add( condition ) ; }

    protected void removeCondition( Condition condition ) {
        conditionList.remove( condition ) ; }

    private void checkConditions() {
        for( int i = 0 ; i < conditionList.size() ; ++i ) {
            Condition condition = conditionList.get(i) ;
            condition.conditionalSignal(); }
    }

    protected State computeNewState() {
        long time = System.currentTimeMillis() ;
        double deltaT = (time - state.time)/1000.0 ;
        double radians = orientationInRadians() ;
        double deltaX = Math.cos( radians ) * state.speed * deltaT ;
        double deltaY = Math.sin( radians ) * state.speed * deltaT ;
        State newState = new State() ;
        newState.position.x = state.position.x + deltaX ;
        newState.position.y = state.position.y + deltaY ;
        newState.orientation = state.orientation + state.angularVelocity * deltaT ;
        newState.speed = state.speed + state.acceleration * deltaT ;
        newState.acceleration = state.acceleration ;
        newState.angularVelocity = state.angularVelocity ;
        newState.time = time ;
        return newState ; }

    protected void updateState( State newState ) {
        keepInBounds( newState ) ;
        avoidCollisions( newState ) ;
        updateHook( newState ) ;
        state = newState ;
        setChanged() ;
        notifyObservers() ;
        checkConditions() ;  }

    protected void keepInBounds( State newState ) {
        if( arena != null ) {
            int width = arena.getWidth() ;
            int height = arena.getHeight() ;
            if( newState.position.x < 0 ) newState.position.x = width ;
            if( newState.position.y < 0 ) newState.position.y = height ;
            if( newState.position.x > width ) newState.position.x = 0 ;
            if( newState.position.y > height ) newState.position.y = 0 ; }
    }

    protected void avoidCollisions( State newState ) {
        if( arena != null ) {
            // Save the old position. Try the new position
            Point2D.Double oldPosition = (Point2D.Double)( state.position.clone() ) ;
            state.position.x = newState.position.x ;
            state.position.y = newState.position.y ;

            boolean collision = false ;
            for( Iterator<Grob> it = arena.iterator() ; !collision && it.hasNext() ; ) {
                Grob next = it.next() ;
                if( next != this && next.isCollidable() && this.intersects( next ) ) {
                    collision = true ; } }

            state.position.x = oldPosition.x ;
            state.position.y = oldPosition.y ;
            if( collision ) {
                newState.position.x = state.position.x ;
                newState.position.y = state.position.y ;
                newState.speed = 0 ; } }
    }


    abstract protected void updateHook( State newState ) ;

    private class UpdateThread extends Thread {
        public void run() {
            while( true ) {
                try {
                    Thread.sleep(10); }
                catch( InterruptedException e ) { }

                monitor.enter() ;
                try {
                    DynamicGrob.State newState = computeNewState() ;
                    updateState(newState) ; }
                finally {
                    monitor.leave() ; } } }
    }

    public void rotateRight(final double delta,  double omega) {
        monitor.enter();
        try {
            final double oldOrientation = state.orientation ;
            final double goal = oldOrientation + delta ;
            State newState = state.copy() ;
            Condition cond ;
            if( delta < 0.0 ) {
                newState.angularVelocity = - Math.abs(omega) ;
                cond = monitor.makeCondition(
                  new Assertion() {
                       public boolean isTrue() {
                           return  state.orientation <= goal ; } } ) ; }
            else {
                newState.angularVelocity = Math.abs(omega) ;
                cond = monitor.makeCondition(
                  new Assertion() {
                       public boolean isTrue() {
                           return  state.orientation >= goal ; } } ) ; }
            addCondition( cond ) ;
            updateState( newState ) ;
            cond.await() ;
            removeCondition( cond ) ;
            // Cheat a bit to correct for overshoot.
                state.orientation = goal ;
            state.angularVelocity = 0.0 ; }
        finally {
            monitor.leave() ; }
    }

    public void rotateLeft( double delta, double omega ) {
        rotateRight( -delta, omega ) ; }

    public void goForward(final double delta,  double speedToGo) {
        monitor.enter();
        try {
            final Point2D.Double oldPosition = (Point2D.Double)( state.position.clone() ) ;
            Condition cond = monitor.makeCondition(
                  new Assertion() {
                       public boolean isTrue() {
                           return state.position.distance(oldPosition) > Math.abs(delta) ; } } ) ;

            State newState = state.copy() ;
            if( delta < 0.0 ) {
                newState.speed = - Math.abs(speedToGo) ; }
            else {
                newState.speed = Math.abs(speedToGo) ; }
            newState.acceleration = 0.0 ;
            newState.angularVelocity = 0.0 ;
            updateState( newState ) ;
            addCondition( cond ) ;
            cond.await() ;
            removeCondition( cond ) ;
            double goalX = oldPosition.x + delta*Math.cos( orientationInRadians() ) ;
            double goalY = oldPosition.y + delta*Math.sin( orientationInRadians() ) ;

            // Fixup any overshoot
            newState = state.copy() ;
            newState.position.x = goalX ;
            newState.position.y = goalY ;
            newState.speed = 0.0 ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public void reset() {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.position = new Point2D.Double(0.0, 0.0) ;
            newState.speed = newState.orientation = newState.acceleration =
              newState.angularVelocity = 0.0 ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }
    
    public void moveBy( double dx, double dy ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.position.x += dx ;
            newState.position.y += dy ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; } }

    public void setPosition( double x, double y ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.position.x = x ;
            newState.position.y = y ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getPositionX() {
        monitor.enter();
        try {
            return state.position.x ; }
        finally {
            monitor.leave() ; }
    }

    public double getPositionY() {
        monitor.enter();
        try {
            return state.position.y ; }
        finally {
            monitor.leave() ; }
    }

    public void setSpeed( double speedToGo ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.speed = speedToGo ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getSpeed() {
        monitor.enter();
        try {
            return state.speed ; }
        finally {
            monitor.leave() ; }
    }

    public void setOrientation( double newOrientation ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.orientation = newOrientation ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public void setDirection( double x, double y ) {
        monitor.enter();
        try {
        	double newOrientation = Math.atan2(y, x) * 180 / Math.PI ;
            State newState = state.copy() ;
            newState.orientation = newOrientation ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getOrientation() {
        monitor.enter();
        try {
            return state.orientation ; }
        finally {
            monitor.leave() ; }
    }

    public void setVelocity( double vx, double vy ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.speed = Math.sqrt(vx*vx + vy*vy) ;
            newState.orientation = Math.atan2(vy, vx) / Math.PI * 180.0 ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getVelocityX() {
        monitor.enter();
        try {
            return state.speed * Math.cos( orientationInRadians() ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getVelocityY() {
        monitor.enter();
        try {
            return state.speed * Math.sin( orientationInRadians() ) ; }
        finally {
            monitor.leave() ; }
    }

    public void setRateOfChangeOfSpeed( double newAcceleration ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.acceleration = newAcceleration ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getRateOfChangeOfSpeed() {
        monitor.enter();
        try {
            return state.acceleration ; }
        finally {
            monitor.leave() ; }
    }

    public void setSpin( double newAngularVelocity ) {
        monitor.enter();
        try {
            State newState = state.copy() ;
            newState.angularVelocity = newAngularVelocity ;
            updateState( newState ) ; }
        finally {
            monitor.leave() ; }
    }

    public double getSpin() {
        monitor.enter();
        try {
            return state.angularVelocity ; }
        finally {
            monitor.leave() ; }
    }

    protected double orientationInRadians() {
        return state.orientation / 180.0 * Math.PI ;
    }

    protected class State {

        /** position is in pixels using screen coordinates.
         * I.e., (0,0) is the right, top corner */
        public Point2D.Double position = new Point2D.Double(0.0, 0.0);

        /** velocity is in pixels per second. */
        public double speed = 0.0 ;

        /** orientation is in degrees. 0 is in the direction of the positive x axis */
        public double orientation = 0.0  ;

        /** acceleration is in pixels per second per second */
        public double acceleration = 0.0 ;

        /** angular velocity is in degrees per second */
        public double angularVelocity = 0.0 ;

        /** time in miliseconds since start of execution */
        public long time ;

        public State copy() {
            State newState = new State() ;
            newState.position = (Point2D.Double) position.clone() ;
            newState.speed = speed ;
            newState.orientation = orientation ;
            newState.acceleration = acceleration ;
            newState.angularVelocity = angularVelocity ;
            newState.time = time ;
            return newState ; }
    }
}